/*

SteamerPackage - CAPE-OPEN steam/water property package
Copyright (C) 2010 Jasper van Baten

You can use this source code provided that
 - its use is for academic teaching or academic research purposes, and
 - its use is non-commercial, and
 - you will not distribute COM objects that use any of the CLSIDs in Steamer.idl
 - you properly acknowledge the author

For any other use, please contact jasper@amsterchem.com

For more information, see LICENSE.txt

*/

#pragma once
#include "resource.h"       // main symbols

//some props
#define MOLWT			18.0153
#define IUPACNAME		L"oxidane"
#define CASNUMBER		L"7732-18-5"

#include "water.h"

#include "Steamer.h"
#include "Variant.h"

#if defined(_WIN32_WCE) && !defined(_CE_DCOM) && !defined(_CE_ALLOW_SINGLE_THREADED_OBJECTS_IN_MTA)
#error "Single-threaded COM objects are not properly supported on Windows CE platform, such as the Windows Mobile platforms that do not include full DCOM support. Define _CE_ALLOW_SINGLE_THREADED_OBJECTS_IN_MTA to force ATL to support creating single-thread COM object's and allow use of it's single-threaded COM object implementations. The threading model in your rgs file was set to 'Free' as that is the only threading model supported in non DCOM Windows CE platforms."
#endif

// CSteamerPackage

class ATL_NO_VTABLE CSteamerPackage :
	public CComObjectRootEx<CComSingleThreadModel>,
	public CComCoClass<CSteamerPackage, &CLSID_SteamerPackage>,
	public IDispatchImpl<ECapeRoot, &__uuidof(ECapeRoot), &LIBID_CAPEOPEN110, /* wMajor = */ 1, /* wMinor = */ 1>,
	public IDispatchImpl<ECapeUnknown, &__uuidof(ECapeUnknown), &LIBID_CAPEOPEN110, /* wMajor = */ 1, /* wMinor = */ 1>,
	public IDispatchImpl<ECapeUser, &__uuidof(ECapeUser), &LIBID_CAPEOPEN110, /* wMajor = */ 1, /* wMinor = */ 1>,
	public IDispatchImpl<ICapeIdentification, &__uuidof(ICapeIdentification), &LIBID_CAPEOPEN110, /* wMajor = */ 1, /* wMinor = */ 1>,
	public IDispatchImpl<ICapeThermoCompounds, &__uuidof(ICapeThermoCompounds), &LIBID_CAPEOPEN110, /* wMajor = */ 1, /* wMinor = */ 1>,
	public IDispatchImpl<ICapeThermoEquilibriumRoutine, &__uuidof(ICapeThermoEquilibriumRoutine), &LIBID_CAPEOPEN110, /* wMajor = */ 1, /* wMinor = */ 1>,
	public IDispatchImpl<ICapeThermoMaterialContext, &__uuidof(ICapeThermoMaterialContext), &LIBID_CAPEOPEN110, /* wMajor = */ 1, /* wMinor = */ 1>,
	public IDispatchImpl<ICapeThermoPhases, &__uuidof(ICapeThermoPhases), &LIBID_CAPEOPEN110, /* wMajor = */ 1, /* wMinor = */ 1>,
	public IDispatchImpl<ICapeThermoPropertyRoutine, &__uuidof(ICapeThermoPropertyRoutine), &LIBID_CAPEOPEN110, /* wMajor = */ 1, /* wMinor = */ 1>,
	public IDispatchImpl<ICapeThermoUniversalConstant, &__uuidof(ICapeThermoPropertyRoutine), &LIBID_CAPEOPEN110, /* wMajor = */ 1, /* wMinor = */ 1>,
	public IDispatchImpl<ICapeUtilities, &__uuidof(ICapeUtilities), &LIBID_CAPEOPEN110, /* wMajor = */ 1, /* wMinor = */ 1>
{
private:
 
   std::wstring name,desc,errName,errDesc,errScope,errIface;
   IDispatch *simContext;
   ICapeThermoMaterial *contextMat;
   double Tboil;
   BSTR temperature,pressure,fraction,phaseFraction,mole,mass,liquid,vapor,enthalpy,entropy;
   
   void SetError(const TCHAR *name,const TCHAR *desc,const TCHAR *iface, const TCHAR *scope)
    {//store error info
     errName=name;
     errDesc=desc;
     errScope=scope;
     errIface=iface;
    }

public:

	CSteamerPackage()
	{name=L"Water";
	 desc=L"Water - IAPWS based Property Package";
	 simContext=NULL;
	 contextMat=NULL;
	 //init normal boiling point
	 Tboil=Water::Tsat(101325.0e-6);
	 //alloc frequently used values:
	 temperature=SysAllocString(L"temperature");
	 pressure=SysAllocString(L"pressure");
	 fraction=SysAllocString(L"fraction");
	 phaseFraction=SysAllocString(L"phaseFraction");
	 mole=SysAllocString(L"mole");
	 mass=SysAllocString(L"mass");
	 liquid=SysAllocString(L"Liquid");
	 vapor=SysAllocString(L"Vapor");
	 enthalpy=SysAllocString(L"enthalpy");
	 entropy=SysAllocString(L"entropy");
	 #ifdef _DEBUG
	 Water w;
	 w.CheckConsistency();
	 #endif
	}
	
	~CSteamerPackage()
	{//call terminate, in case it has not yet been called by the COSE
	 Terminate();
	 //done with string values
	 SysFreeString(temperature);
	 SysFreeString(pressure);
	 SysFreeString(fraction);
	 SysFreeString(phaseFraction);
	 SysFreeString(mole);
	 SysFreeString(mass);
	 SysFreeString(liquid);
	 SysFreeString(vapor);
	 SysFreeString(enthalpy);
	 SysFreeString(entropy);
	}

	DECLARE_REGISTRY_RESOURCEID(IDR_STEAMERPACKAGE)

	BEGIN_COM_MAP(CSteamerPackage)
		COM_INTERFACE_ENTRY2(IDispatch, ICapeThermoPropertyRoutine)
		COM_INTERFACE_ENTRY(ECapeRoot)
		COM_INTERFACE_ENTRY(ECapeUnknown)
		COM_INTERFACE_ENTRY(ECapeUser)
		COM_INTERFACE_ENTRY(ICapeIdentification)
		COM_INTERFACE_ENTRY(ICapeThermoCompounds)
		COM_INTERFACE_ENTRY(ICapeThermoEquilibriumRoutine)
		COM_INTERFACE_ENTRY(ICapeThermoMaterialContext)
		COM_INTERFACE_ENTRY(ICapeThermoPhases)
		COM_INTERFACE_ENTRY(ICapeThermoPropertyRoutine)
		COM_INTERFACE_ENTRY(ICapeThermoUniversalConstant)
		COM_INTERFACE_ENTRY(ICapeUtilities)
	END_COM_MAP()

	DECLARE_PROTECT_FINAL_CONSTRUCT()

	HRESULT FinalConstruct()
	{
		return S_OK;
	}

	void FinalRelease()
	{
	}
	
	BOOL GetSinglePhaseScalarProp(double &val,BSTR prop,BSTR phase,BSTR basis,std::wstring &err)
	{//check before hand that we have a context material
	 ATLASSERT(contextMat);
	 std::wstring ws;
	 VARIANT v;
	 HRESULT hr;
	 VariantInit(&v);
	 hr=contextMat->GetSinglePhaseProp(prop,phase,basis,&v);
	 if (FAILED(hr))
	  {err=L"Failed to get ";
	   err+=prop;
	   err+=L" for phase ";
	   err+=phase;
	   err+=L": ";
	   err+=CO_Error(contextMat,hr);
	   return FALSE;
	  }
	 CVariant propVal(v,TRUE);
	 if (!propVal.CheckArray(VT_R8,ws))
	  {err=L"Invalid value for ";
	   err+=prop;
	   err+=L" for phase ";
	   err+=phase;
	   err+=L": ";
	   err+=ws;
	   return FALSE;
	  }
	 if (propVal.GetCount()!=1)
	  {err=L"Invalid value for ";
	   err+=prop;
	   err+=L" for phase ";
	   err+=phase;
	   err+=L": scalar expected";
	   return FALSE;
	  }
	 val=propVal.GetDoubleAt(0);
	 return TRUE;
	}

	BOOL GetOverallScalarProp(double &val,BSTR prop,BSTR basis,std::wstring &err)
	{//check before hand that we have a context material
	 ATLASSERT(contextMat);
	 std::wstring ws;
	 VARIANT v;
	 HRESULT hr;
	 VariantInit(&v);
	 hr=contextMat->GetOverallProp(prop,basis,&v);
	 if (FAILED(hr))
	  {err=L"Failed to get ";
	   err+=prop;
	   err+=L" for overall phase: ";
	   err+=CO_Error(contextMat,hr);
	   return FALSE;
	  }
	 CVariant propVal(v,TRUE);
	 if (!propVal.CheckArray(VT_R8,ws))
	  {err=L"Invalid value for ";
	   err+=prop;
	   err+=L" for overall phase: ";
	   err+=ws;
	   return FALSE;
	  }
	 if (propVal.GetCount()!=1)
	  {err=L"Invalid value for ";
	   err+=prop;
	   err+=L" for overall phase: ";
	   return FALSE;
	  }
	 val=propVal.GetDoubleAt(0);
	 return TRUE;
	}

	BOOL SetSinglePhaseScalarProp(double val,BSTR prop,BSTR phase,BSTR basis,std::wstring &err)
	{//check before hand that we have a context material
	 ATLASSERT(contextMat);
	 std::wstring ws;
	 HRESULT hr;
	 CVariant v;
	 v.MakeArray(1,VT_R8);
	 v.SetDoubleAt(0,val);
	 hr=contextMat->SetSinglePhaseProp(prop,phase,basis,v.Value());
	 if (FAILED(hr))
	  {err=L"Failed to set ";
	   err+=prop;
	   err+=L" for phase ";
	   err+=phase;
	   err+=L": ";
	   err+=CO_Error(contextMat,hr);
	   return FALSE;
	  }
	 return TRUE;
	}

	BOOL SetOverallScalarProp(double val,BSTR prop,BSTR basis,std::wstring &err)
	{//check before hand that we have a context material
	 ATLASSERT(contextMat);
	 std::wstring ws;
	 HRESULT hr;
	 CVariant v;
	 v.MakeArray(1,VT_R8);
	 v.SetDoubleAt(0,val);
	 hr=contextMat->SetOverallProp(prop,basis,v.Value());
	 if (FAILED(hr))
	  {err=L"Failed to set ";
	   err+=prop;
	   err+=L" for overall phase: ";
	   err+=CO_Error(contextMat,hr);
	   return FALSE;
	  }
	 return TRUE;
	}

public:


	// ECapeRoot Methods

	STDMETHOD(get_name)(BSTR * name)
	{	//error name
		*name=SysAllocString(errName.c_str());
		return NO_ERROR;
	}

	// ECapeUser Methods

	STDMETHOD(get_code)(long * code)
	{	//not supported
	    *code=0;
		return NO_ERROR;
	}
	
	STDMETHOD(get_description)(BSTR * description)
	{	//error description  
	    *description=SysAllocString(errDesc.c_str());	 
		return NO_ERROR;
	}
	
	STDMETHOD(get_scope)(BSTR * scope)
	{	//error scope
	    *scope=SysAllocString(errScope.c_str());
		return NO_ERROR;
	}
	
	STDMETHOD(get_interfaceName)(BSTR * interfaceName)
	{	//error interface
	    *interfaceName=SysAllocString(errIface.c_str());
		return NO_ERROR;
	}
	
	STDMETHOD(get_operation)(BSTR * operation)
	{	//error operation - not supported
	    *operation=SysAllocString(L"N/A");
		return NO_ERROR;
	}
	
	STDMETHOD(get_moreInfo)(BSTR * moreInfo)
	{	//more info, not supported
	    *moreInfo=SysAllocString(L"N/A");
		return NO_ERROR;
	}

	// ICapeIdentification Methods
public:

	STDMETHOD(get_ComponentName)(BSTR * name)
	{	//object name
	    *name=SysAllocString(this->name.c_str());
		return NO_ERROR;
	}
	
	STDMETHOD(put_ComponentName)(BSTR name)
	{	//set name 
	    if (name) this->name=name;
	    else this->name.clear();
		return NO_ERROR;
	}
	
	STDMETHOD(get_ComponentDescription)(BSTR * desc)
	{	//component description
	    *desc=SysAllocString(this->desc.c_str());
		return NO_ERROR;
	}
	
	STDMETHOD(put_ComponentDescription)(BSTR desc)
	{	//set desc
	    if (desc) this->desc=desc;
	    else this->desc.clear();
		return NO_ERROR;
	}

	// ICapeThermoCompounds Methods
public:

	STDMETHOD(GetCompoundConstant)(VARIANT props, VARIANT compIds, VARIANT * propVals)
	{int compCount=0;
	 int i,j,index;
	 BSTR b;
	 std::wstring error;
	 //check compound list
	 if (compIds.vt==VT_EMPTY) compCount=1; 
	 else
	  {CVariant compList(compIds,FALSE);
	   if (!compList.CheckArray(VT_BSTR,error))
	    {error=L"Invalid compound list: "+error;
	     SetError(L"Invalid compound list",error.c_str(),L"ICapeThermoCompounds",L"GetCompoundConstant");
	     return ECapeUnknownHR;
	    }
	   if (compList.GetCount()==0)
	    {//assume one compound
	     compCount=1;
	    }
	   else
	    {for (i=0;i<compList.GetCount();i++)
	      {b=compList.GetStringAt(i);
	       if (!SameString(b,L"Water"))
	        {error=L"Invalid compound in compound list: ";
	         error+=b;
	         SysFreeString(b);
	         SetError(L"Invalid compound",error.c_str(),L"ICapeThermoCompounds",L"GetCompoundConstant");
	         return ECapeUnknownHR;
	        }
	       SysFreeString(b);
	      }
	     compCount=compList.GetCount();
	    }
	  }
	 //check property list
	 CVariant propList(props,FALSE);
	 if (!propList.CheckArray(VT_BSTR,error))
      {error=L"Invalid property list: "+error;
       SetError(L"Invalid property list",error.c_str(),L"ICapeThermoCompounds",L"GetCompoundConstant");
       return ECapeUnknownHR;
      }
     if (propList.GetCount()==0)
      {//done
       VariantClear(propVals);
       return NO_ERROR;
      }
     //prep result array
     CVariant res;
     res.MakeArray(compCount*propList.GetCount(),VT_VARIANT);
     //loop over prop list
     index=0;
     for (i=0;i<propList.GetCount();i++)
      {VARIANT v;
       b=propList.GetStringAt(i);
       if (SameString(b,L"criticalDensity")) //mol/m3
        {v.vt=VT_R8;
         v.dblVal=RHOCRIT/(MOLWT*1e-3);
        }
       else if (SameString(b,L"criticalPressure"))
        {v.vt=VT_R8;
         v.dblVal=PCRIT*1e6;
        }
       else if (SameString(b,L"criticalTemperature"))
        {v.vt=VT_R8;
         v.dblVal=TCRIT;
        }
       else if (SameString(b,L"criticalVolume"))
        {v.vt=VT_R8;
         v.dblVal=MOLWT*1e-3/RHOCRIT;
        }
       else if (SameString(b,L"molecularWeight"))
        {v.vt=VT_R8;
         v.dblVal=MOLWT;
        }
       else if (SameString(b,L"triplePointPressure"))
        {v.vt=VT_R8;
         v.dblVal=PTRIP*1e6;
        }
       else if (SameString(b,L"triplePointTemperature"))
        {v.vt=VT_R8;
         v.dblVal=TTRIP;
        }
       else if (SameString(b,L"casRegistryNumber"))
        {v.vt=VT_BSTR;
         v.bstrVal=SysAllocString(CASNUMBER);
        }
       else if (SameString(b,L"chemicalFormula"))
        {v.vt=VT_BSTR;
         v.bstrVal=SysAllocString(L"H2O");
        }
       else if (SameString(b,L"iupacName"))
        {v.vt=VT_BSTR;
         v.bstrVal=SysAllocString(IUPACNAME);
        }
       else if (SameString(b,L"normalBoilingPoint"))
        {v.vt=VT_R8;
         v.dblVal=Tboil;
        }
       else
        {//not supported
         error=L"Compound constant property \"";
         error+=b;
         SysFreeString(b);
         error+=L"\" is not supported";
         SetError(L"Property not supported",error.c_str(),L"ICapeThermoCompounds",L"GetCompoundConstant");
         return ECapeUnknownHR;
        }
       for (j=0;j<compCount;j++)
        {res.SetVariantAt(index,v);
         index++;
        }
       SysFreeString(b);
       VariantClear(&v);
      }
     *propVals=res.ReturnValue();
	 return NO_ERROR;
	}

	STDMETHOD(GetCompoundList)(VARIANT * compIds, VARIANT * formulae, VARIANT * names, VARIANT * boilTemps, VARIANT * molwts, VARIANT * casnos)
	{	//all arguments are [in, out];
	    VariantClear(compIds);
	    VariantClear(formulae);
	    VariantClear(names);
	    VariantClear(boilTemps);
	    VariantClear(molwts);
	    VariantClear(casnos);
	 	//only water
	 	CVariant list;
	 	BSTR b;
	 	list.MakeArray(1,VT_BSTR);
	 	b=SysAllocString(L"Water");
	 	list.SetStringAt(0,b);
	 	SysFreeString(b);
	 	*compIds=list.Copy();
	 	*names=list.ReturnValue();
	 	list.MakeArray(1,VT_BSTR);
	 	b=SysAllocString(L"H2O");
	 	list.SetStringAt(0,b);
	 	SysFreeString(b);
	 	*formulae=list.ReturnValue();
	 	list.MakeArray(1,VT_R8);
	 	list.SetDoubleAt(0,Tboil);
	 	*boilTemps=list.ReturnValue();
	 	list.MakeArray(1,VT_R8);
	 	list.SetDoubleAt(0,MOLWT);
	 	*molwts=list.ReturnValue();
	 	list.MakeArray(1,VT_BSTR);
	 	b=SysAllocString(CASNUMBER);
	 	list.SetStringAt(0,b);
	 	SysFreeString(b);
	 	*casnos=list.ReturnValue();
		return NO_ERROR;
	}
	
	STDMETHOD(GetConstPropList)(VARIANT * props)
	{	CVariant list;
	    BSTR b;
	    list.MakeArray(11,VT_BSTR);
	    b=SysAllocString(L"criticalDensity");list.SetStringAt(0,b);SysFreeString(b);
	    b=SysAllocString(L"criticalPressure");list.SetStringAt(1,b);SysFreeString(b);
	    b=SysAllocString(L"criticalTemperature");list.SetStringAt(2,b);SysFreeString(b);
	    b=SysAllocString(L"criticalVolume");list.SetStringAt(3,b);SysFreeString(b);
	    b=SysAllocString(L"molecularWeight");list.SetStringAt(4,b);SysFreeString(b);
	    b=SysAllocString(L"triplePointPressure");list.SetStringAt(5,b);SysFreeString(b);
	    b=SysAllocString(L"triplePointTemperature");list.SetStringAt(6,b);SysFreeString(b);
	    b=SysAllocString(L"casRegistryNumber");list.SetStringAt(7,b);SysFreeString(b);
	    b=SysAllocString(L"chemicalFormula");list.SetStringAt(8,b);SysFreeString(b);
	    b=SysAllocString(L"iupacName");list.SetStringAt(9,b);SysFreeString(b);
	    b=SysAllocString(L"normalBoilingPoint");list.SetStringAt(10,b);SysFreeString(b);
	    *props=list.ReturnValue();
		return NO_ERROR;
	}
	
	STDMETHOD(GetNumCompounds)(long * num)
	{   //just water
	    *num=1;
		return NO_ERROR;
	}
	
	STDMETHOD(GetPDependentProperty)(VARIANT props, double pressure, VARIANT compIds, VARIANT * propVals)
	{	//no P dependent properties
	    SetError(L"Not supported",L"This package does not expose P-Dependent properties",L"ICapeThermoCompounds",L"GetPDependentProperty");
		return ECapeUnknownHR;
	}
	
	STDMETHOD(GetPDependentPropList)(VARIANT * props)
	{	//return empty list
	    props->vt=VT_EMPTY;
		return NO_ERROR;
	}
	
	STDMETHOD(GetTDependentProperty)(VARIANT props, double temperature, VARIANT compIds, VARIANT * propVals)
	{int compCount=0;
	 int i,j,index;
	 BSTR b;
	 std::wstring error;
	 //check valid T
	 if (isMissing(temperature)) 
	  {SetError(L"Missing temperature",L"Temperature not specified or not finite",L"ICapeThermoCompounds",L"GetTDependentProperty");
	   return ECapeUnknownHR;
	  }
	 if (temperature<0)
	  {SetError(L"Zero or negative temperature",L"Zero or negative temperature",L"ICapeThermoCompounds",L"GetTDependentProperty");
	   return ECapeUnknownHR;
	  }
	 //check compound list
	 if (compIds.vt==VT_EMPTY) compCount=1; 
	 else
	  {CVariant compList(compIds,FALSE);
	   if (!compList.CheckArray(VT_BSTR,error))
	    {error=L"Invalid compound list: "+error;
	     SetError(L"Invalid compound list",error.c_str(),L"ICapeThermoCompounds",L"GetTDependentProperty");
	     return ECapeUnknownHR;
	    }
	   if (compList.GetCount()==0)
	    {//assume one compound
	     compCount=1;
	    }
	   else
	    {for (i=0;i<compList.GetCount();i++)
	      {b=compList.GetStringAt(i);
	       if (!SameString(b,L"Water"))
	        {error=L"Invalid compound in compound list: ";
	         error+=b;
	         SysFreeString(b);
	         SetError(L"Invalid compound",error.c_str(),L"ICapeThermoCompounds",L"GetTDependentProperty");
	         return ECapeUnknownHR;
	        }
	       SysFreeString(b);
	      }
	     compCount=compList.GetCount();
	    }
	  }
	 //check property list
	 CVariant propList(props,FALSE);
	 if (!propList.CheckArray(VT_BSTR,error))
      {error=L"Invalid property list: "+error;
       SetError(L"Invalid property list",error.c_str(),L"ICapeThermoCompounds",L"GetTDependentProperty");
       return ECapeUnknownHR;
      }
     if (propList.GetCount()==0)
      {//done
       VariantClear(propVals);
       return NO_ERROR;
      }
     //prep result array
     CVariant res;
     res.MakeArray(compCount*propList.GetCount(),VT_R8);
     //loop over prop list
     index=0;
     for (i=0;i<propList.GetCount();i++)
      {double value;
       b=propList.GetStringAt(i);
       try
        {if (SameString(b,L"surfaceTensionSatLiquid")) 
          {value=Water::surfaceTension(temperature);
          }
         else if (SameString(b,L"thermalConductivityOfLiquid"))
          {double Psat=Water::Psat(temperature);
           Water w;
           w.SetStateTPX(temperature,Psat*1e-6,TPLIQUID);
           value=w.thermalConductivity();
          }
         else if (SameString(b,L"thermalConductivityOfVapor"))
          {double Psat=Water::Psat(temperature);
           Water w;
           w.SetStateTPX(temperature,Psat*1e-6,TPVAPOR);
           value=w.thermalConductivity();
          }
         else if (SameString(b,L"vaporPressure"))
          {value=Water::Psat(temperature)*1e6;
          }
         else if (SameString(b,L"volumeChangeUponVaporization"))
          {double vVap;
           double vLiq;
           double Psat=Water::Psat(temperature);
           Water w;
           w.SetStateTPX(temperature,Psat*1e-6,TPLIQUID);
           vLiq=w.volume();
           w.SetStateTPX(temperature,Psat*1e-6,TPVAPOR);
           vVap=w.volume();
		   value=(vVap-vLiq)*(MOLWT*1e-3); //m3/mol
          }
         else if (SameString(b,L"volumeOfLiquid"))
          {double Psat=Water::Psat(temperature);
           Water w;
           w.SetStateTPX(temperature,Psat*1e-6,TPLIQUID);
           value=w.volume()*(MOLWT*1e-3);//m3/mol
          }
         else if (SameString(b,L"viscosityOfLiquid"))
          {double Psat=Water::Psat(temperature);
           Water w;
           w.SetStateTPX(temperature,Psat*1e-6,TPLIQUID);
           value=w.viscosity();
          }
         else if (SameString(b,L"viscosityOfVapor"))
          {double Psat=Water::Psat(temperature);
           Water w;
           w.SetStateTPX(temperature,Psat*1e-6,TPVAPOR);
           value=w.viscosity();
          }
         else if (SameString(b,L"idealGasEnthalpy"))
          {//we return enthalpy at P=0
           //this property is just provided for reference state calculations for ideal gas reference state
           Water w;
           w.SetStateTPX(temperature,0,TPVAPOR);
           value=w.enthalpy()*MOLWT;//J/kg
          }
         else if (SameString(b,L"idealGasEntropy"))
          {//use external water ideal heat capacity correlation
           //this property is just provided for reference state calculations for ideal gas reference state
           const double A=36542.0;
           const double B=-34.8040008545;
           const double C=0.116820000112;
           const double D=-0.000130040003569;
           const double E=5.25470014168e-008;
           const double Pref=101325e-6;
           const double Tref=373.15; //vapor exists at Tref, Pref
           //RPP ideal gas heat capacity = A+B*T+C*T^2+D*T^3+E*T^4
           //ideal gas entropy: integral from Tref to T over Cp/T
           //parameters from Properties of Gases and Liquids 5th Edition, by Prausnitz
           value=temperature*(B+temperature*(0.5*C+temperature*(D/3.0+temperature*0.25*E)))+A*log(temperature)-
                (Tref*(B+Tref*(0.5*C+Tref*(D/3.0+Tref*0.25*E)))+A*log(Tref));
           value*=1e-3;
           //add the reference state value:
           Water w;
           w.SetStateTPX(Tref,Pref,TPVAPOR);
           value+=w.entropy()*MOLWT;//J/kg/K
          }
         else
          {//not supported
           error=L"T-dependent property \"";
           error+=b;
           SysFreeString(b);
           error+=L"\" is not supported";
           SetError(L"Property not supported",error.c_str(),L"ICapeThermoCompounds",L"GetTDependentProperty");
           return ECapeUnknownHR;
          }
        }
       catch(char *str)
        {error=L"Calculation error for property ";
         error+=b;
         SysFreeString(b);
         error+=L": ";
         error+=CA2CT(str);
         SetError(L"Calculation error",error.c_str(),L"ICapeThermoCompounds",L"GetTDependentProperty");
         return ECapeUnknownHR; 
        }
       for (j=0;j<compCount;j++)
        {res.SetDoubleAt(index,value);
         index++;
        }
       SysFreeString(b);
      }
     *propVals=res.ReturnValue();
	 return NO_ERROR;
	}
	
	STDMETHOD(GetTDependentPropList)(VARIANT * props)
	{   CVariant list;
	    BSTR b;
	    list.MakeArray(10,VT_BSTR);
	    b=SysAllocString(L"surfaceTensionSatLiquid");list.SetStringAt(0,b);SysFreeString(b);
	    b=SysAllocString(L"thermalConductivityOfLiquid");list.SetStringAt(1,b);SysFreeString(b);
	    b=SysAllocString(L"thermalConductivityOfVapor");list.SetStringAt(2,b);SysFreeString(b);
	    b=SysAllocString(L"vaporPressure");list.SetStringAt(3,b);SysFreeString(b);
	    b=SysAllocString(L"volumeChangeUponVaporization");list.SetStringAt(4,b);SysFreeString(b);
	    b=SysAllocString(L"volumeOfLiquid");list.SetStringAt(5,b);SysFreeString(b);
	    b=SysAllocString(L"viscosityOfLiquid");list.SetStringAt(6,b);SysFreeString(b);
	    b=SysAllocString(L"viscosityOfVapor");list.SetStringAt(7,b);SysFreeString(b);
	    b=SysAllocString(L"idealGasEntropy");list.SetStringAt(8,b);SysFreeString(b);
	    b=SysAllocString(L"idealGasEnthalpy");list.SetStringAt(9,b);SysFreeString(b);
	    *props=list.ReturnValue();
		return NO_ERROR;
	}

	// ICapeThermoPhases Methods

	STDMETHOD(GetNumPhases)(long * num)
	{	//liquid and vapor
	    *num=2;
		return NO_ERROR;
	}

	STDMETHOD(GetPhaseInfo)(BSTR phaseLabel, BSTR phaseAttribute, VARIANT * value)
	{	BOOL vapor=FALSE;
	    //check phase label
	    if (SameString(phaseLabel,L"Vapor")) vapor=TRUE;
	    else if (SameString(phaseLabel,L"Liquid")) vapor=FALSE;
	    else
	     {//invalid
	      std::wstring error;
	      error=L"Invalid phase label: ";
	      error+=(phaseLabel)?phaseLabel:L"NULL";
	      SetError(L"Invalid phase label",error.c_str(),L"ICapeThermoPhases",L"GetPhaseInfo");
	      return ECapeUnknownHR;
	     }
	    //check phaseAttribute
	    value->vt=VT_BSTR;
	    value->bstrVal=NULL;
	    if (SameString(phaseAttribute,L"StateOfAggregation")) value->bstrVal=SysAllocString((vapor)?L"Vapor":L"Liquid");
	    else if (SameString(phaseAttribute,L"KeyCompoundId")) value->bstrVal=SysAllocString(L"Water");
	    else if ((SameString(phaseAttribute,L"ExcludedCompoundId"))|| 
	             (SameString(phaseAttribute,L"DensityDescription"))|| 
	             (SameString(phaseAttribute,L"UserDescription"))|| 
	             (SameString(phaseAttribute,L"TypeOfSolid"))) {}
	    else
	     {//not valid 
	      std::wstring error;
	      error=L"Invalid phase attribute: ";
	      error+=(phaseAttribute)?phaseAttribute:L"NULL";
	      SetError(L"Invalid phase attribute",error.c_str(),L"ICapeThermoPhases",L"GetPhaseInfo");
	      return ECapeUnknownHR;
	     }
		return NO_ERROR;
	}

	STDMETHOD(GetPhaseList)(VARIANT * phaseLabels, VARIANT * stateOfAggregation, VARIANT * keyCompoundId)
	{	//all arguments are [in, out]
	    VariantClear(phaseLabels);
	    VariantClear(stateOfAggregation);
	    VariantClear(keyCompoundId);
	    //return lists
	    CVariant list;
	    BSTR b;
	    list.MakeArray(2,VT_BSTR);
	    b=SysAllocString(L"Vapor");
	    list.SetStringAt(0,b);
	    SysFreeString(b);
	    b=SysAllocString(L"Liquid");
	    list.SetStringAt(1,b);
	    SysFreeString(b);
	    *phaseLabels=list.Copy();
	    *stateOfAggregation=list.ReturnValue();
	    b=SysAllocString(L"Water");
	    list.MakeArray(2,VT_BSTR);
	    list.SetStringAt(0,b);
	    list.SetStringAt(1,b);
	    SysFreeString(b);
	    *keyCompoundId=list.ReturnValue();
		return NO_ERROR;
	}

	// ICapeThermoPropertyRoutine Methods

	STDMETHOD(CalcAndGetLnPhi)(BSTR phaseLabel, double temperature, double pressure, VARIANT moleNumbers, int fFlags, VARIANT * lnPhi, VARIANT * lnPhiDT, VARIANT * lnPhiDP, VARIANT * lnPhiDn)
	{	//we have no fugacity 
	    SetError(L"Not implemented",L"Fugacity coefficients are not implemented for this package",L"ICapeThermoPropertyRoutine",L"CalcAndGetLnPhi");
		return ECapeNoImplHR;
	}
	
	STDMETHOD(GetSinglePhasePropList)(VARIANT * props)
	{   CVariant list;
	    BSTR b;
	    list.MakeArray(11,VT_BSTR);
	    b=SysAllocString(L"density");list.SetStringAt(0,b);SysFreeString(b);
	    b=SysAllocString(L"enthalpy");list.SetStringAt(1,b);SysFreeString(b);
	    b=SysAllocString(L"entropy");list.SetStringAt(2,b);SysFreeString(b);
	    b=SysAllocString(L"heatCapacityCp");list.SetStringAt(3,b);SysFreeString(b);
	    b=SysAllocString(L"heatCapacityCv");list.SetStringAt(4,b);SysFreeString(b);
	    b=SysAllocString(L"internalEnergy");list.SetStringAt(5,b);SysFreeString(b);
	    b=SysAllocString(L"molecularWeight");list.SetStringAt(6,b);SysFreeString(b);
	    b=SysAllocString(L"thermalConductivity");list.SetStringAt(7,b);SysFreeString(b);
	    b=SysAllocString(L"volume");list.SetStringAt(8,b);SysFreeString(b);
	    b=SysAllocString(L"viscosity");list.SetStringAt(9,b);SysFreeString(b);
	    b=SysAllocString(L"gibbsEnergy");list.SetStringAt(10,b);SysFreeString(b);
	    *props=list.ReturnValue();
		return NO_ERROR;
	}

	STDMETHOD(CheckSinglePhasePropSpec)(BSTR property, BSTR phaseLabel, VARIANT_BOOL * valid)
	{	//check if valid prop spec
	 	*valid=VARIANT_TRUE;
	 	//check phase label
	 	if ((!SameString(phaseLabel,L"Vapor"))&&(!SameString(phaseLabel,L"Liquid")))
	 	 {//not valid
	 	  *valid=VARIANT_FALSE;
	 	 }
	 	else
	 	 {//check property 
	 	  if ((!SameString(property,L"density"))&&
	 	      (!SameString(property,L"enthalpy"))&&
	 	      (!SameString(property,L"entropy"))&&
	 	      (!SameString(property,L"gibbsEnergy"))&&
	 	      (!SameString(property,L"heatCapacityCp"))&&
	 	      (!SameString(property,L"heatCapacityCv"))&&
	 	      (!SameString(property,L"internalEnergy"))&&
	 	      (!SameString(property,L"molecularWeight"))&&
	 	      (!SameString(property,L"thermalConductivity"))&&
	 	      (!SameString(property,L"volume"))&&
	 	      (!SameString(property,L"viscosity")))
	 	   {//not valid
	 	    *valid=VARIANT_FALSE;
	 	   }
	 	 }
		return NO_ERROR;
	}

	STDMETHOD(CalcSinglePhaseProp)(VARIANT props, BSTR phaseLabel)
	{	BOOL isSteam;
	    std::wstring err;
	    double T,P,X;
	    int i;
	    std::vector<BSTR> bases;
	    std::vector<double> values;
	 	//check context material
		if (!contextMat)
		 {SetError(L"No context material",L"Context material object not set",L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		  return ECapeUnknownHR;
		 }
		//check phase
		if (SameString(phaseLabel,L"Vapor")) isSteam=TRUE;
		else if (SameString(phaseLabel,L"Liquid")) isSteam=FALSE;
		else
		 {//invalid phase label
		  err=L"Invalid phase label: ";
		  err+=(phaseLabel)?phaseLabel:L"NULL";
		  SetError(L"Invalid phase label",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		  return ECapeUnknownHR;
		 }
		//check prop list
		CVariant pList(props,FALSE);
		if (!pList.CheckArray(VT_BSTR,err))
		 {err=L"Invalid property list: "+err;
		  SetError(L"Invalid property list",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		  return ECapeUnknownHR;
		 }
		//get temperature
		if (!GetSinglePhaseScalarProp(T,temperature,phaseLabel,NULL,err))
		 {SetError(L"Failed to get property",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		  return ECapeUnknownHR;
		 }
		//get pressure
		if (!GetSinglePhaseScalarProp(P,pressure,phaseLabel,NULL,err))
		 {SetError(L"Failed to get property",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		  return ECapeUnknownHR;
		 }
		//get composition
		if (!GetSinglePhaseScalarProp(X,fraction,phaseLabel,mole,err))
		 {SetError(L"Failed to get property",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		  return ECapeUnknownHR;
		 }
		//check valid composition
		if (fabs(X-1.0)>1e-3)
		 {SetError(L"Invalid composition",L"Invalid composition: not unity",L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		  return ECapeUnknownHR;
		 }
		//check valid T
		if (isMissing(T))
		 {SetError(L"Invalid temperature",L"Missing or non-finite value for temperature",L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		  return ECapeUnknownHR;
		 }
		//check valid P
		if (isMissing(P))
		 {SetError(L"Invalid pressure",L"Missing or non-finite value for pressure",L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		  return ECapeUnknownHR;
		 }
		//alloc space
		bases.resize(pList.GetCount());
		values.resize(pList.GetCount());
		//prep calculator
		Water w;
		try
		 {w.SetStateTPX(T,P*1e-6,(isSteam)?TPVAPOR:TPLIQUID);
		 }
		catch (char *str)
		 {err=L"Failed to set state: ";
		  err+=CA2CT(str);
		  SetError(L"Invalid state",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		  return ECapeUnknownHR;
		 }
		//loop over all properties to calculate:
		for (i=0;i<pList.GetCount();i++)
		 {BSTR propName=pList.GetStringAt(i);
		  try
		   {if (SameString(propName,L"density"))
		     {values[i]=1.0/(w.volume());//kg/m3
		      bases[i]=mass;
		     }
		    else if (SameString(propName,L"enthalpy"))
		     {values[i]=w.enthalpy()*1e3; //J/kg
		      bases[i]=mass;
		     }
		    else if (SameString(propName,L"entropy"))
		     {values[i]=w.entropy()*1e3; //J/kg/K
		      bases[i]=mass;
		     }
		    else if (SameString(propName,L"gibbsEnergy"))
		     {values[i]=w.gibbs()*1e3; //J/kg
		      bases[i]=mass;
		     }
		    else if (SameString(propName,L"heatCapacityCp"))
		     {values[i]=w.cP()*1e3;//J/kg/K
		      bases[i]=mass;
		     }
		    else if (SameString(propName,L"heatCapacityCv"))
		     {values[i]=w.cV()*1e3;//J/kg/K
		      bases[i]=mass;
		     }
		    else if (SameString(propName,L"internalEnergy"))
		     {values[i]=w.energy()*1e3; //J/kg
		      bases[i]=mass;
		     }
		    else if (SameString(propName,L"molecularWeight")) 
		     {values[i]=MOLWT; //relative
		      bases[i]=NULL;
		     }
		    else if (SameString(propName,L"thermalConductivity"))
		     {values[i]=w.thermalConductivity();//W/m/K
		      bases[i]=NULL;
		     }
		    else if (SameString(propName,L"volume"))
		     {values[i]=w.volume();//m3/kg
		      bases[i]=mass;
		     }
		    else if (SameString(propName,L"viscosity"))
		     {values[i]=w.viscosity();//Pa.s
		     }
		    else
		     {//invalid / unsupported property
		      err=L"Invalid or unsupported single-phase property ";
		      err+=(propName)?propName:L"NULL";
		      SysFreeString(propName);
		      SetError(L"Invalid property",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		      return ECapeUnknownHR;		   
		     }
		   }
		  catch (char *str)
		   {err=L"Failed to calculate ";
		    err+=(propName)?propName:L"NULL";
		    SysFreeString(propName);
		    err+=L": ";
		    err+=CA2CT(str);
		    SetError(L"Invalid state",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		    return ECapeUnknownHR;
		   }
		  SysFreeString(propName);
		 }
		//all properties have been calculated, set all properties on material object
		for (i=0;i<pList.GetCount();i++)
		 {BSTR propName=pList.GetStringAt(i);
		  if (!SetSinglePhaseScalarProp(values[i],propName,phaseLabel,bases[i],err))
		   {SysFreeString(propName);
		    SetError(L"Failed to set property",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcSinglePhaseProp");
		    return ECapeUnknownHR;
		   }
		  SysFreeString(propName);
		 }
		return NO_ERROR;
	}

	STDMETHOD(GetTwoPhasePropList)(VARIANT * props)
	{	CVariant list;
	    BSTR b;
	    list.MakeArray(1,VT_BSTR);
	    b=SysAllocString(L"surfaceTension");list.SetStringAt(0,b);SysFreeString(b);
	    *props=list.ReturnValue();
		return NO_ERROR;
	}

	STDMETHOD(CheckTwoPhasePropSpec)(BSTR property, VARIANT phaseLabels, VARIANT_BOOL * valid)
	{	std::wstring err;
		int i;
		BOOL haveVap,haveLiq;
		//check if valid prop spec
	 	*valid=VARIANT_TRUE;
	 	//check phase labels
	 	CVariant phases(phaseLabels,FALSE);
	 	if (!phases.CheckArray(VT_BSTR,err))
	 	 {//invalid phase list 
	 	  err=L"Invalid phase list: "+err;
	 	  SetError(L"Invalid phase list",err.c_str(),L"ICapeThermoPropertyRoutine",L"CheckTwoPhasePropSpec");
	 	  return ECapeUnknownHR;
	 	 }
	 	if (phases.GetCount()!=2)
	 	 {//invalid phase list 
	 	  SetError(L"Invalid phase list",L"Invalid phase list: expected two phase labels",L"ICapeThermoPropertyRoutine",L"CheckTwoPhasePropSpec");
	 	  return ECapeUnknownHR;
	 	 }
	 	//check that phases are vapor and liquid
	 	haveVap=haveLiq=FALSE;
	 	for (i=0;i<2;i++)
	 	 {BSTR b=phases.GetStringAt(i);
	 	  if (SameString(b,L"Vapor")) haveVap=TRUE;
	 	  else if (SameString(b,L"Liquid")) haveLiq=TRUE;
	 	  SysFreeString(b);
	 	 }
	 	if ((!haveVap)||(!haveLiq))
	 	 {//not valid
	 	  *valid=VARIANT_FALSE;
	 	 }
	 	else
	 	 {//check property 
	 	  if (!SameString(property,L"surfaceTension"))
	 	   {//not valid
	 	    *valid=VARIANT_FALSE;
	 	   }
	 	 }
		return NO_ERROR;
	}

	STDMETHOD(CalcTwoPhaseProp)(VARIANT props, VARIANT phaseLabels)
	{   std::wstring err;
		int i;
		double T;
		BOOL haveVap,haveLiq;
		BSTR propName;
	 	//check context material
		if (!contextMat)
		 {SetError(L"No context material",L"Context material object not set",L"ICapeThermoPropertyRoutine",L"CalcTwoPhaseProp");
		  return ECapeUnknownHR;
		 }	 	
	 	//check phase labels
	 	CVariant phases(phaseLabels,FALSE);
	 	if (!phases.CheckArray(VT_BSTR,err))
	 	 {//invalid phase list 
	 	  err=L"Invalid phase list: "+err;
	 	  SetError(L"Invalid phase list",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcTwoPhaseProp");
	 	  return ECapeUnknownHR;
	 	 }
	 	if (phases.GetCount()!=2)
	 	 {//invalid phase list 
	 	  SetError(L"Invalid phase list",L"Invalid phase list: expected two phase labels",L"ICapeThermoPropertyRoutine",L"CalcTwoPhaseProp");
	 	  return ECapeUnknownHR;
	 	 }
	 	//check that phases are vapor and liquid
	 	haveVap=haveLiq=FALSE;
	 	for (i=0;i<2;i++)
	 	 {BSTR b=phases.GetStringAt(i);
	 	  if (SameString(b,L"Vapor")) haveVap=TRUE;
	 	  else if (SameString(b,L"Liquid")) haveLiq=TRUE;
	 	  else
	 	   {err=L"Invalid phase: ";
	 	    err+=(b)?b:L"NULL";
	 	    SysFreeString(b);
	 	    SetError(L"Invalid phase",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcTwoPhaseProp");
	 	    return ECapeUnknownHR;
	 	   }
	 	  SysFreeString(b);
	 	 }
	 	if ((!haveVap)||(!haveLiq))
	 	 {//not valid
	 	  SetError(L"Unsupported phase combination",L"Two-phase properties are only supported for the Vapor plus Liquid phase.",L"ICapeThermoPropertyRoutine",L"CalcTwoPhaseProp");
	 	  return ECapeUnknownHR;
	 	 }
		//check prop list
		CVariant pList(props,FALSE);
		if (!pList.CheckArray(VT_BSTR,err))
		 {err=L"Invalid property list: "+err;
		  SetError(L"Invalid property list",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcTwoPhaseProp");
		  return ECapeUnknownHR;
		 }
		if (pList.GetCount()==0)
		 {//done
		  return NO_ERROR;
		 }
		//get liquid temperature
		if (!GetSinglePhaseScalarProp(T,temperature,liquid,NULL,err))
		 {SetError(L"Failed to get property",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcTwoPhaseProp");
		  return ECapeUnknownHR;
		 }
		//check valid T
		if (isMissing(T))
		 {SetError(L"Invalid temperature",L"Missing or non-finite value for temperature",L"ICapeThermoPropertyRoutine",L"CalcTwoPhaseProp");
		  return ECapeUnknownHR;
		 }
	    //loop over properties
	    for (i=0;i<pList.GetCount();i++)
	     {propName=pList.GetStringAt(i);
	      if (!SameString(propName,L"surfaceTension"))
	       {//not supported
	        err=L"Invalid or unsupported property: ";
	        err+=(propName)?propName:L"NULL";
	        SysFreeString(propName);
	        SetError(L"Invalid or unsupported property",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcTwoPhaseProp");
	        return ECapeUnknownHR;
	       }
	      SysFreeString(propName);
	     }
	    //calc surface tension
	    double st;
	    try 
	     {st=Water::surfaceTension(T);
	     }
	    catch (char *str)
	     {err=L"Calculation error: ";
	      err+=CA2CT(str);
	      SetError(L"Calculation error",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcTwoPhaseProp");
	      return ECapeUnknownHR;
	     }
	    //set results on material object
	    CVariant vals;
	    vals.MakeArray(1,VT_R8);
	    vals.SetDoubleAt(0,st);
	    propName=SysAllocString(L"surfaceTension");
	    HRESULT hr=contextMat->SetTwoPhaseProp(propName,phaseLabels,NULL,vals.Value());
	    SysFreeString(propName);
	    if (FAILED(hr))
	     {err=L"Failed to set surfaceTension on context material: ";
	      err+=CO_Error(contextMat,hr);
	      SetError(L"Failed to set property",err.c_str(),L"ICapeThermoPropertyRoutine",L"CalcTwoPhaseProp");
	      return ECapeUnknownHR;
	     }
		return NO_ERROR;
	}

	// ICapeThermoEquilibriumRoutine Methods
	
	enum EqType
	 {//most commonly used flashes, and the ones supported by this object
	  TP,	//pressure, temperature
	  TVF,	//temperature, vapor fraction
	  PVF,	//pressure, vapor fraction
	  PH,	//pressure, enthalpy
	  TH,	//temperature, enthalpy
	  PS,	//pressure, entropy
	  TS, 	//temperature, entropy
	  HVF,  //enthalpy, vapor fraction
	  SVF   //entropy, vapor fraction
	 };
	 
	BOOL GetEquilibriumType(VARIANT specification1, VARIANT specification2, BSTR solutionType, EqType &type, BSTR &VFBasis,BOOL &vfVapor, std::wstring &error)
	{//this routine does most of the work of CheckEquilibriumSpec and part of the work of CalcEquilibrium
	 int i;
	 HRESULT hr;
	 BOOL haveVap,haveLiq;
	 BSTR b,prop,base,phase;
	 std::wstring err;
	 BOOL haveT,haveP,haveVF,haveH,haveS;
	 VFBasis=NULL; //fill in if vapor fraction is part of flash type
	 vfVapor=TRUE; //set to false for liquid fraction spec
	 haveT=haveP=haveVF=haveH=haveS=FALSE;
	 //check if we have a context material
	 if (!contextMat)
	  {error=L"Context material not set";
	   return FALSE;
	  }
	 //get list of present phases (we only support flashes that can result vapor and liquid)
	 VARIANT labels,status;
	 VariantInit(&labels);
	 VariantInit(&status);
	 hr=contextMat->GetPresentPhases(&labels,&status);
	 if (FAILED(hr)) 
	  {error=L"Failed to get list of present phases from context material: ";
	   error+=CO_Error(contextMat,hr);
	   return FALSE;
	  }
	 VariantClear(&status); //we do not care about phase status or initial guess
	 CVariant phaseList(labels,TRUE);
	 if (!phaseList.CheckArray(VT_BSTR,err))  
	  {error=L"Invalid list of present phases from context material: ";
	   error+=err;
	   return FALSE;
	  }
	 if (phaseList.GetCount()!=2)
	  {error=L"Only flashes that can result vapor and liquid are supported";
	   return FALSE;
	  }
	 haveVap=haveLiq=FALSE;
	 for (i=0;i<2;i++)
	  {BSTR label=phaseList.GetStringAt(i);
	   if (SameString(label,vapor)) haveVap=TRUE;
	   else if (SameString(label,liquid)) haveLiq=TRUE;
	  }
	 if ((!haveVap)||(!haveLiq))
	  {error=L"Only flashes that can result vapor and liquid are supported";
	   return FALSE;
	  }
	 //check both specifications, to see what they are
	 for (i=0;i<2;i++)
	  {CVariant spec((i)?specification2:specification1,FALSE);
	   //check spec format
	   prop=base=phase=NULL;
	   if (!spec.CheckArray(VT_BSTR,err))
	    {//set err before jumping here:
	     specError:
	     error=L"Invalid specification ";
	     error+=(i)?L'2':L'1';
	     error+=L": ";
	     error+=err;
	     if (prop) SysFreeString(prop);
	     if (base) SysFreeString(base);
	     if (phase) SysFreeString(phase);
	     return FALSE;
	    }
	   if ((spec.GetCount()<3)||(spec.GetCount()>4))
	    {err=L"expected 3 or 4 elements";
	     goto specError;
	    }
	   //get property, basis and phase
	   prop=spec.GetStringAt(0);
	   base=spec.GetStringAt(1);
	   phase=spec.GetStringAt(2);
	   //which property?
	   if (SameString(prop,temperature))
	    {//T spec, only for overall
	     if (!SameString(phase,L"Overall"))
	      {err=L"temperature specification only supported for overall phase";
	       goto specError;
	      }
	     if (!SameString(base,NULL))
	      {err=L"unsupported basis for temperature";
	       goto specError;
	      }
	     if (haveT)
	      {err=L"Temperature-temperature flash not supported";
	       goto specError;
	      }
	     haveT=TRUE;
	    }
	   else if (SameString(prop,pressure))
	    {//P spec, only for overall
	     if (!SameString(phase,L"Overall"))
	      {err=L"pressure specification only supported for overall phase";
	       goto specError;
	      }
	     if (!SameString(base,NULL))
	      {err=L"unsupported basis for pressure";
	       goto specError;
	      }
	     if (haveP)
	      {err=L"Pressure-pressure flash not supported";
	       goto specError;
	      }
	     haveP=TRUE;
	    }
	   else if (SameString(prop,phaseFraction))
	    {//phase fraction spec
	     if (SameString(phase,L"Overall"))
	      {err=L"overall phase fraction not supported";
	       goto specError;
	      }
	     if (SameString(phase,liquid)) vfVapor=FALSE;
	     else if (!SameString(phase,vapor))
	      {err=L"invalid phase for phase fraction specification";
	       goto specError;
	      }
	     if (SameString(base,mass)) VFBasis=mass;
	     else if (SameString(base,mole)) VFBasis=mole;
	     else
	      {err=L"unsupported basis for phase fraction specification";
	       goto specError;
	      }
	     if (haveVF)
	      {err=L"phase fraction - phase fraction flash not supported";
	       goto specError;
	      }
	     haveVF=TRUE;
	    }
	   else if (SameString(prop,L"enthalpy"))
	    {//H spec, only for overall
	     if (!SameString(phase,L"Overall"))
	      {err=L"enthalpy specification only supported for overall phase";
	       goto specError;
	      }
	     if (haveH)
	      {err=L"Enthalpy-enthalpy flash not supported";
	       goto specError;
	      }
	     haveH=TRUE;
	    }
	   else if (SameString(prop,L"entropy"))
	    {//S spec, only for overall
	     if (!SameString(phase,L"Overall"))
	      {err=L"entropy specification only supported for overall phase";
	       goto specError;
	      }
	     if (haveS)
	      {err=L"Entropy-entropyflash not supported";
	       goto specError;
	      }
	     haveS=TRUE;
	    }
	   else 
	    {err=L"Unsupported property ";
	     err+=(prop)?prop:L"NULL";
	     goto specError;
	    }
       //we do not support a flash spec for a certain compound, just for the whole mixture of only water	   
	   if (spec.GetCount()==4)
	    {b=spec.GetStringAt(3);
	     if (!SameString(b,NULL))
	      {err=L"compound specification not supported";
	       SysFreeString(b);
	       goto specError;
	      }
	     SysFreeString(b);
	    }
	   //done with spec string
	   SysFreeString(prop);
	   SysFreeString(phase);
	   SysFreeString(base);
	  }
	 //check flash type
	 if (haveT)
	  {if (haveP) type=TP;
	   else if (haveVF) type=TVF;
	   else if (haveH) type=TH;
	   else if (haveS) type=TS;
	   else 
	    {unsupportedCombination:
	     error=L"Unsupported combination of flash specification variables";
	     return FALSE;
	    }
	  }
	 else if (haveP)
	  {if (haveVF) type=PVF;
	   else if (haveH) type=PH;
	   else if (haveS) type=PS;
	   else goto unsupportedCombination;
	  }
	 else if (haveVF)
	  {if (haveH) type=HVF;
	   else if (haveS) type=SVF;
	   else goto unsupportedCombination;
	  }
	 else goto unsupportedCombination;
	 //check solutionType
	 if ((!SameString(solutionType,NULL))&&(!SameString(solutionType,L"Unspecified")))
	  {//a solution type is specified
	   if (SameString(solutionType,L"Normal"))
	    {//normal, only ok if VF is specified
	     if (!haveVF) 
	      {error=L"Solution type \"normal\" is only supported for phase fraction flashes";
	       return FALSE;
	      }
	    }
	   else if (SameString(solutionType,L"Retrograde"))
	    {error=L"Retrograde flashes are not supported";
	     return FALSE;
	    }
	   else
	    {error=L"Invalid or unsupported solution type: ";
	     error+=solutionType;
	     return FALSE;
	    }
	  }
	 //flash spec is ok and determined
	 return TRUE;
	}

	STDMETHOD(CalcEquilibrium)(VARIANT specification1, VARIANT specification2, BSTR solutionType)
	{	EqType type;
	    BOOL vfVapor;
	    BSTR VFBasis;
	    std::wstring error;
	    BOOL haveT,haveP,haveVF;
	    double T,P,X,H,S;
	    //get equilibrium calculation spec
	    if (!GetEquilibriumType(specification1,specification2,solutionType,type,VFBasis,vfVapor,error))
	     {//not supported or error in spec:
	      SetError(L"Unsupported equilibrium spec",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
	      return ECapeUnknownHR;
	     }
	    //check overall composition
        if (!GetOverallScalarProp(X,fraction,mole,error))
         {SetError(L"Failed to get overall composition",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
          return ECapeUnknownHR;
         }
	    if (fabs(X-1.0)>1e-3)
	     {SetError(L"Invalid overall composition",L"Overall composition is not unity",L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
          return ECapeUnknownHR;
	     }
	    //get inputs, dependent on flash type
	    haveT=haveP=haveVF=FALSE;
	    switch (type)
	     {case TP:
	      case TVF:
	      case TH:
	      case TS:
	        //get overall temperature
	        if (!GetOverallScalarProp(T,temperature,NULL,error))
	         {SetError(L"Failed to get temperature",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
	          return ECapeUnknownHR;
	         }
			//check valid T
			if (isMissing(T)) 
			 {SetError(L"Missing temperature",L"Temperature not specified or not finite",L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
			  return ECapeUnknownHR;
			 }
			if (T<0)
			 {SetError(L"Zero or negative temperature",L"Zero or negative temperature",L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
			  return ECapeUnknownHR;
			 }
			//T ok 
			haveT=TRUE;
		}
	    switch (type)
	     {case TP:
	      case PVF:
	      case PH:
	      case PS:
	        //get overall pressure
	        if (!GetOverallScalarProp(P,pressure,NULL,error))
	         {SetError(L"Failed to get pressure",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
	          return ECapeUnknownHR;
	         }
			//check valid P
			if (isMissing(P)) 
			 {SetError(L"Missing pressure",L"Pressure not specified or not finite",L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
			  return ECapeUnknownHR;
			 }
			//P ok 
			haveP=TRUE;
		}
	    switch (type)
	     {case TVF:
	      case PVF:
	      case HVF:
	      case SVF:
	         if (vfVapor)
	          {//get vapor fraction
	           if (!GetSinglePhaseScalarProp(X,phaseFraction,vapor,VFBasis,error))
	            {SetError(L"Failed to get vapor fraction",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
	             return ECapeUnknownHR;
	            }
	          }
	         else
	          {//get liquid fraction
	           if (!GetSinglePhaseScalarProp(X,phaseFraction,liquid,VFBasis,error))
	            {SetError(L"Failed to get liquid fraction",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
	             return ECapeUnknownHR;
	            }
	           X=1.0-X; //make vapor fraction
	          }
	         if (isMissing(X))
	          {SetError(L"Missing phase fraction",L"Phase fraction not specified or not finite",L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
			   return ECapeUnknownHR;
			  }
			 if ((X<-.0000001)||(X>1.0000001))
			  {//out of range
			   SetError(L"Invalid phase fraction",L"Phase fraction out of [0,1] range",L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
			   return ECapeUnknownHR;
			  }
			 //VF ok 
			 haveVF=TRUE;
		}
	    switch (type)
	     {case TH:
	      case PH:
	      case HVF:
	        //get overall enthalpy
	        if (!GetOverallScalarProp(H,enthalpy,mass,error))
	         {SetError(L"Failed to get enthalpy",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
	          return ECapeUnknownHR;
	         }
			//check valid H
			if (isMissing(H)) 
			 {SetError(L"Missing enthalpy",L"Enthalpy not specified or not finite",L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
			  return ECapeUnknownHR;
			 }
		}
	    switch (type)
	     {case TS:
	      case PS:
	      case SVF:
	        //get overall entropy
	        if (!GetOverallScalarProp(S,entropy,mass,error))
	         {SetError(L"Failed to get entropy",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
	          return ECapeUnknownHR;
	         }
			//check valid S
			if (isMissing(S)) 
			 {SetError(L"Missing entropy",L"Entropy not specified or not finite",L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
			  return ECapeUnknownHR;
			 }
		}
	   //calculate the flash, at the end we need valid T, P, X(=vapor fraction)
	   try
	    {switch (type)
	      {case TP:	//pressure, temperature
	            {Water w;
	             w.SetStateTPX(T,P*1e-6,TPAUTO);
	             X=w.GetQuality();
	            }
	            break;
	       case TVF: //temperature, vapor fraction
	            P=Water::Psat(T)*1e6;
	            break;
	       case PVF:	//pressure, vapor fraction
	            T=Water::Tsat(P*1e-6);
	            break;
	       case PH:	//pressure, enthalpy
	            {//check two-phase solution (not well dealt with by solver)
	             Water w;
	             w.SetStatePH(P*1e-6,H*1e-3);
	             T=w.GetTemperature();
	             X=w.GetQuality();
	            }
	            break;
	       case TH:	//temperature, enthalpy
	            {Water w;
	             w.SetStateTH(T,H*1e-3);
	             P=w.GetPressure()*1e6;
	             X=w.GetQuality();
	            }
	            break;
	       case PS:	//pressure, entropy
	            {Water w;
	             w.SetStatePS(P*1e-6,S*1e-3);
	             T=w.GetTemperature();
	             X=w.GetQuality();
	            }
	           break;
	       case TS:	//temperature, entropy
	            {Water w;
	             w.SetStateTS(T,S*1e-3);
	             P=w.GetPressure()*1e6;
	             X=w.GetQuality();
	            }
	           break;
	       case HVF:	//enthalpy, vapor fraction
	            {Water w;
	             w.SetStateHVF(H*1e-3,X);
	             P=w.GetPressure()*1e6;
	             T=w.GetTemperature();
	            }
	           break;
	       case SVF:	//entropy, vapor fraction
	            {Water w;
	             w.SetStateSVF(S*1e-3,X);
	             P=w.GetPressure()*1e6;
	             T=w.GetTemperature();
	            }
	           break;
	      }
	    }
	   catch (char *str)
	    {error=L"Flash calculation error: ";
	     error+=CA2CT(str);
	     SetError(L"Calculation error",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");
	     return ECapeUnknownHR;
	    }
	   //flash calculation ok, set result on context material
	   BOOL haveVap,haveLiq;
	   int phaseCount,index;
	   haveVap=((X>0)||(haveVF));
	   haveLiq=((X<1.0)||(haveVF));
	   //set present phases
	   phaseCount=0;
	   if (haveVap) phaseCount++;
	   if (haveLiq) phaseCount++;
	   CVariant phaseList,statList;
	   phaseList.MakeArray(phaseCount,VT_BSTR);
	   index=0;
	   if (haveVap) phaseList.SetStringAt(index++,vapor);
	   if (haveLiq) phaseList.SetStringAt(index++,liquid);
	   statList.MakeArray(phaseCount,VT_I4);
	   for (index=0;index<phaseCount;index++) statList.SetLongAt(index,CAPE_ATEQUILIBRIUM);
	   HRESULT hr=contextMat->SetPresentPhases(phaseList.Value(),statList.Value());
	   if (FAILED(hr))
	    {error=L"Failed to set present phases: ";
	     error+=CO_Error(contextMat,hr);
	     SetError(L"Failed to set present phases",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");	     
	     return ECapeUnknownHR;
	    }
	   if (X<0) X=0;
	   else if (X>1.0) X=1.0;
	   if (haveVap)
	    {//set vapor VF, T, P, X 
	     if (!SetSinglePhaseScalarProp(X,phaseFraction,vapor,mole,error))
	      {SetError(L"Failed to set vapor fraction",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");	     
	       return ECapeUnknownHR;
	      }
	     if (!SetSinglePhaseScalarProp(T,temperature,vapor,NULL,error))
	      {SetError(L"Failed to set vapor temperature",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");	     
	       return ECapeUnknownHR;
	      }
	     if (!SetSinglePhaseScalarProp(P,pressure,vapor,NULL,error))
	      {SetError(L"Failed to set vapor pressure",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");	     
	       return ECapeUnknownHR;
	      }
	     if (!SetSinglePhaseScalarProp(1.0,fraction,vapor,mole,error))
	      {SetError(L"Failed to set vapor composition",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");	     
	       return ECapeUnknownHR;
	      }
	    }
	   if (haveLiq)
	    {//set liquid VF, T, P, X 
	     if (!SetSinglePhaseScalarProp(1.0-X,phaseFraction,liquid,mole,error))
	      {SetError(L"Failed to set liquid fraction",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");	     
	       return ECapeUnknownHR;
	      }
	     if (!SetSinglePhaseScalarProp(T,temperature,liquid,NULL,error))
	      {SetError(L"Failed to set liquid temperature",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");	     
	       return ECapeUnknownHR;
	      }
	     if (!SetSinglePhaseScalarProp(P,pressure,liquid,NULL,error))
	      {SetError(L"Failed to set liquid pressure",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");	     
	       return ECapeUnknownHR;
	      }
	     if (!SetSinglePhaseScalarProp(1.0,fraction,liquid,mole,error))
	      {SetError(L"Failed to set liquid composition",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");	     
	       return ECapeUnknownHR;
	      }
	    }
	   if (!haveT)
	    {//set overall T as it was not part of the flash spec
	     if (!SetOverallScalarProp(T,temperature,NULL,error))
	      {SetError(L"Failed to set overall temperature",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");	     
	       return ECapeUnknownHR;
	      }
	    }
	   if (!haveP)
	    {//set overall P as it was not part of the flash spec
	     if (!SetOverallScalarProp(P,pressure,NULL,error))
	      {SetError(L"Failed to set overall pressure",error.c_str(),L"ICapeThermoEquilibriumRoutine",L"CalcEquilibrium");	     
	       return ECapeUnknownHR;
	      }
	    }
	   return NO_ERROR;
	}

	STDMETHOD(CheckEquilibriumSpec)(VARIANT specification1, VARIANT specification2, BSTR solutionType, VARIANT_BOOL * isSupported)
	{	//check support for equilibrium calculation 
	    EqType type;
	    BOOL vfVapor;
	    BSTR VFBasis;
	    std::wstring error;
	 	*isSupported=VARIANT_TRUE;
	    if (!GetEquilibriumType(specification1,specification2,solutionType,type,VFBasis,vfVapor,error))
	     {//not supported
	      *isSupported=VARIANT_FALSE;
	     }
		return NO_ERROR;
	}

	// ICapeThermoMaterialContext Methods

	STDMETHOD(SetMaterial)(LPDISPATCH material)
	{	std::wstring err;
	 	if (contextMat)
	     {contextMat->Release();
	      contextMat=NULL;
	     }
	    if (material)
	     {//get material interface
	      HRESULT hr;
	      hr=material->QueryInterface(IID_ICapeThermoMaterial,(LPVOID*)&contextMat);
	      if (FAILED(hr))
	       {err=L"Failed to get ICapeThermoMaterial: ";
	        err+=HResError(hr);
	        SetError(L"No material interface",err.c_str(),L"ICapeThermoMaterialContext",L"SetMaterial");
	        contextMat=NULL;
	        return ECapeUnknownHR;
	       }
	      //check that the material object only contains water
	      ICapeThermoCompounds *cmpds;
	      hr=contextMat->QueryInterface(IID_ICapeThermoCompounds,(LPVOID*)&cmpds);
	      if (FAILED(hr))
	       {err=L"Failed to get ICapeThermoCompounds from material: ";
	        err+=HResError(hr);
	        SetError(L"No compounds interface",err.c_str(),L"ICapeThermoMaterialContext",L"SetMaterial");
	        contextMat->Release();
	        contextMat=NULL;
	        return ECapeUnknownHR;
	       }
	      VARIANT compIds,formulae,names,boiltemps,molwts,casnos;
	      VariantInit(&compIds);
	      VariantInit(&formulae);
	      VariantInit(&names);
	      VariantInit(&boiltemps);
	      VariantInit(&molwts);
	      VariantInit(&casnos);
	      hr=cmpds->GetCompoundList(&compIds,&formulae,&names,&boiltemps,&molwts,&casnos);
	      if (FAILED(hr))
	       {err=L"Failed to get list of compounds from material: ";
	        err+=CO_Error(cmpds,hr);
	        SetError(L"Failed to get compounds",err.c_str(),L"ICapeThermoMaterialContext",L"SetMaterial");
	        cmpds->Release();
	        contextMat->Release();
	        contextMat=NULL;
	        return ECapeUnknownHR;
	       }
	      cmpds->Release();
	      //we do not use these
	      VariantClear(&formulae);
	      VariantClear(&names);
	      VariantClear(&boiltemps);
	      VariantClear(&molwts);
	      VariantClear(&casnos);
	      //check the compound ID list
	      CVariant idList(compIds,TRUE); //idList now owns compIds, do not destroy
	      if (!idList.CheckArray(VT_BSTR,err))
	       {err=L"Invalid list of compound IDs from material object: "+err;
	        SetError(L"Invalid compound list",err.c_str(),L"ICapeThermoMaterialContext",L"SetMaterial");
	        contextMat->Release();
	        contextMat=NULL;
	        return ECapeUnknownHR;
	       }
	      if (idList.GetCount()!=1)
	       {//we can not have a sub-set of 1 compounds, so we expect exactly one compound
	        SetError(L"Invalid compound list",L"Invalid compound list: expected material containing only water",L"ICapeThermoMaterialContext",L"SetMaterial");
	        contextMat->Release();
	        contextMat=NULL;
	        return ECapeUnknownHR;
	       }
	      BSTR b=idList.GetStringAt(0);
	      if (!SameString(b,L"Water"))
	       {SetError(L"Invalid compound list",L"Invalid compound list: expected material containing only water",L"ICapeThermoMaterialContext",L"SetMaterial");
	        contextMat->Release();
	        contextMat=NULL;
	        SysFreeString(b);
	        return ECapeUnknownHR;
	       }
	      SysFreeString(b);
	     }
		return NO_ERROR;
	}

	STDMETHOD(UnsetMaterial)()
	{   if (contextMat)
	     {contextMat->Release();
	      contextMat=NULL;
	     }
		return NO_ERROR;
	}

	// ICapeUtilities Methods

	STDMETHOD(get_parameters)(LPDISPATCH * parameters)
	{	//not supported 
	    *parameters=NULL;
	    SetError(L"No parameters",L"This object does not expose parameters",L"ICapeUtilities",L"get_parameters");
		return ECapeUnknownHR;
	}
	
	STDMETHOD(put_simulationContext)(LPDISPATCH context)
	{	if (simContext) simContext->Release();
	    if (context)
	     {//keep reference to sim context 
	      simContext=context;
	      context->AddRef();
	     }
	    else simContext=NULL; //should not happen, not allowed to pass NULL pointers in COM
		return NO_ERROR;
	}
	
	STDMETHOD(Initialize)()
	{	//nothing to do
		return NO_ERROR;
	}
	
	STDMETHOD(Terminate)()
	{	//release sim context
	    if (simContext)
	     {simContext->Release();
	      simContext=NULL;
	     }
	    if (contextMat)
	     {contextMat->Release();
	      contextMat=NULL;
	     }
		return NO_ERROR;
	}
	
	STDMETHOD(Edit)()
	{	//not supported
		SetError(L"Not implemented",L"Edit is not implemented by this object",L"ICapeUtilities",L"Edit");	    
		return ECapeNoImplHR;
	}

	// ICapeThermoUniversalContant Methods

	STDMETHOD(GetUniversalConstant)(BSTR constantId, VARIANT * constantValue)
	{   if (SameString(constantId,L"molarGasConstant"))
		 {constantValue->vt=VT_R8;
		  constantValue->dblVal=8.31447215;
		 }
		else if (SameString(constantId,L"avogadroConstant"))
		 {constantValue->vt=VT_R8;
		  constantValue->dblVal=6.0221419947e23;
		 }
		else if (SameString(constantId,L"boltzmannConstant"))
		 {constantValue->vt=VT_R8;
		  constantValue->dblVal=1.380650324e-23;
		 }
		else if (SameString(constantId,L"speedOfLightInVacuum"))
		 {constantValue->vt=VT_R8;
		  constantValue->dblVal=299792458.1;
		 }
		else if (SameString(constantId,L"standardAccelerationOfGravity "))
		 {constantValue->vt=VT_R8;
		  constantValue->dblVal=9.80665;
		 }
		else
		 {//unknown 
		  SetError(L"Unknown constant ID",L"Unknown or unsupported universal constant",L"ICapeThermoUniversalContants",L"GetUniversalConstant");
		  return ECapeUnknownHR;
		 }
		return NOERROR;
	}

	STDMETHOD (GetUniversalConstantList)(VARIANT * constantId)
	{   CVariant list;
	    BSTR b;
	    list.MakeArray(5,VT_BSTR);
	    b=SysAllocString(L"molarGasConstant");list.SetStringAt(0,b);SysFreeString(b);
	    b=SysAllocString(L"avogadroConstant");list.SetStringAt(1,b);SysFreeString(b);
	    b=SysAllocString(L"boltzmannConstant");list.SetStringAt(2,b);SysFreeString(b);
	    b=SysAllocString(L"speedOfLightInVacuum");list.SetStringAt(3,b);SysFreeString(b);
	    b=SysAllocString(L"standardAccelerationOfGravity");list.SetStringAt(4,b);SysFreeString(b);
	    *constantId=list.ReturnValue();
		return NO_ERROR;
	}

	
};

OBJECT_ENTRY_AUTO(__uuidof(SteamerPackage), CSteamerPackage)
